;;; algorithmus-von-prim-gallenbachergraph.scm

(load "gallenbacher-graph-kanten.scm")
; Verwendet wird die Kantenliste "*kanten*" mit der Zugriffsfunktion
; (nachfolge-kanten aktuelle-stadt kanten)
;;; --------------------------------------------------------------------
;;; ***** "Beginne mit einem Baum, der allein aus einer Knoten besteht."
; Der Parameter kuerzeste ist die Liste der bereits gefundenen Kanten.
; s.u.

;;; ***** "Ordne alle Kanten von diesem Knoten aus nach ihren Kosten. 
;;; (→ Prioritätswarteschlange nach den Bewertungen)"
;;; Zwei Teilfunktionen
;;; 1: gibt zu einem Knoten alle Kanten geordnet nach ihren Kosten zurueck
;;; 2. fuegt diese in die PrioWS ein.
;;;;; -----Bewertung-----
;;; sinnvoller:
;;; 1: gibt zu einem Knoten alle Kanten zurueck
;;; 2. fuegt diese [an richtiger Position] in die PrioWS ein.

;;; ===== PrioWS =====
(load "prioritaetswarteschlange.scm")
; stellt die Funktionen (fuege-einen-ein und) fuege-alle-ein bereit
; benötigt ein Praedikat vor?
(define
  (vor? kante-1 kante-2)
  (< (third kante-1) (third kante-2)))
; Tests:
;(vor? '(D C 6) '(A B 10)) ; -> #t
;(vor? '(D C 6) '(D E 3)) ; -> #f

;;; ***** "Füge nun jeweils immer aus dieser jeweils die nächste Kante mit den geringsten Kosten ein,
;;;        die keinen Zyklus erzeugt."
; Praedikat wird benoetigt:
; Es testet, ob beide Knoten in der Knotenliste enthalten sind.
(define (zyklus? kante besucht)
   (member (second kante) besucht))
; Test:
;(zyklus? '(B A 10) '(B A))         ; #t oder etwas, das nicht #f ist
;(zyklus? '(F E 10) '(G F B H E D)) ; #t oder etwas, das nicht #f ist
;(zyklus? '(B H 10) '(B A))         ; #f

;;; ***** Dann füge alle von dem freien Knoten ausgehenden zulässigen Kanten in die Prioritätswarteschlange ein."
; dazu fehlt das Herausfinden des freien Knoten
(define
  (freier-knoten kante)
  (second kante))
; Test:
;(freie-knoten '(A B 10) '(B C D)) ; -> 'A

;;; ***** Brich damit ab, wenn alle Knoten des Graphen zum Baum gehören."
;;; rekursiv besser formuliert mit: "Rufe dich selbst auf, wenn ... nicht ..."
; Praedikat wird benoetigt:
(define
  (alle-enthalten? besucht anzahl)
  (= (length besucht) anzahl))
; Test:
;(alle-enthalten? '(A B) 8)             ; #f
;(alle-enthalten? '(A B C D E F G H) 8) ; #t

;;; ===== Hilfsfunktion anzahl-knoten =====
; Voraussetzung: Die Liste kanten enthaelt jeweils
; auch die Rueckkante
(define
  (anzahl-knoten kanten)
  (define
    (anzahl-intern kanten gefunden)
    ; arbeitet endrekursiv
    (cond
      ((null? kanten) (length gefunden))
      ((member (first (first kanten)) gefunden)
       (anzahl-intern (rest kanten) gefunden))
      (else
       (anzahl-intern
        (rest kanten)
        (cons (first (first kanten)) gefunden)))))
  (anzahl-intern kanten '()))

;;; ----- Funktionen zur Uebersichtlichkeit des Aufrufs
; fuegt den freien Knoten der neuen Kante der Besuchtliste hinzu
(define
  (aktualisiere-besuchtliste knoten besucht)
  (cond
   ((member knoten besucht)
    besucht)
   (else
    (cons knoten besucht))
   ))
; Test:
; (aktualisiere-besuchtliste '(C H 8) '(C E D)) ; oder
; (aktualisiere-besuchtliste 'H '(C E D))

(define
  (kante-in-baum  kante kuerzeste)
  (cons kante kuerzeste))
; Test:
;(kante-in-baum '(C H 8) '((E C 4) (D E 3)))

(define
  (aktualisiere-priows priows kanten)
  (fuege-alle-ein
   (nachfolge-kanten (second (first priows)) kanten)
   vor?
   (rest priows)))


;;; ===== prim =====
; mit einer passenden Aufrufhuelle
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    (cond

      [(alle-enthalten? besucht anzahl)
       kuerzeste]
    
      [(null? priows)
       #f]
    
      [(zyklus? (first priows) besucht) ; freier Knoten in besucht?
       (prim-intern besucht anzahl kuerzeste (rest priows))]
    
      [else
       (prim-intern
        (aktualisiere-besuchtliste (freier-knoten (first priows)) besucht)
        anzahl
        (kante-in-baum (first priows) kuerzeste)
        (aktualisiere-priows priows kanten))]
      )) ; Ende interne Funktion
  ; interner Aufruf:
  (prim-intern
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   ))

;Test:
(prim 'I *kanten*)