;;; prim-kanten-Erarbeitung.scm

; Das Programm benoetigt die ausgegliederten Funktionen:
(load "prioritaetswarteschlange.scm")
; mit der Funktion (fuege-alle-ein <kantenliste> vor? <priows>),
; die intern die Funktion (fuege-eine-ein <kante> vor? <priows>) verwendet
; Das Praedikat vor? muss dafuer definiert werden:
(define
  (vor? kante-1 kante-2)
  (< (third kante-1) (third kante-2)))
; und
(load "gallenbacher-graph-kanten.scm")
; Verwendet wird davon die Liste "kanten"
; verwenden Sie einen anderen Graphen muss auch vorhanden sein:
; die Funktion (nachfolge-kanten <aktuelle-stadt> <kanten>)

;;; =============== Entwurfsprinzip fuer Prim ===============
;;; Zum Augangs- und letzten Zielknoten werden die Nachfolgekanten
;;; in die Prioritaetswarteschlange (priows) eingefuegt
;;; bis alle Knoten besucht worden sind

;;; ===== prim =====
; mit einer passenden Aufrufhuelle
; ---------------
; Schritt 0: Funktionskopf
(define
  (prim start-knoten kanten)
  'fehlt)

; ---------------
; Schritt 1: Funktionskopf der internen Funktion einbauen
; (prim-intern <besuchte Knoten> <anzahl-knoten> <kuerzeste-kanten> <kanten> <priows>)
; und deren Aufruf mit den richtigen Startparametern
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    'fehlt)                              ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   'anfangswert-besuchtliste
   'anzahl-knoten
   'liste-der-kuerzesten-wege ; ist noch leer
   'nachfolgekanten-vom-startknoten-geordnet)
  )
; anfangswert-besuchtliste: (list  
; anzahl-knoten Hilfsfunktion aufrufen: (anzahl-knoten kanten)
; liste-der-kuerzesten-wege ist noch leer: '()
; nachfolgekanten-vom-startknoten-geordnet
; Nachfolgekanten vom Startknoten holen (Hilfsfunktion)
; und in die leere Priows einordnen

;;; ===== Hilfsfunktion anzahl-knoten =====
; Voraussetzung: Die Liste kanten enthaelt jeweils
; auch die Rueckkante
(define
  (anzahl-knoten kanten)
  (define
    (anzahl-intern kanten gefunden)
    ; arbeitet endrekursiv
    (cond
      ((null? kanten) (length gefunden))
      ((member (first (first kanten)) gefunden)
       (anzahl-intern (rest kanten) gefunden))
      (else
       (anzahl-intern
        (rest kanten)
        (cons (first (first kanten)) gefunden)))))
  (anzahl-intern kanten '()))
       
; ----- Schritt 1 umgesetzt:
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    'fehlt)                              ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   )
  )

; ---------------
; Schritt 2:
; Der elementare Abbruchfall: 
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    (cond
      ((null? priows)
       #f)
      (else
       'fehlt)
      ))                                 ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   )
  )
;;; ===== Tests:
;(prim 'O '((X Y 2)))
;(prim 'O kanten)

; ---------------
; Schritt 3:
; Das Ziel ist erreicht: Alle Knoten wurden erreicht
; Erfolgsfall ohne ZUgriff auf priows, also vorher testen
; -> Liste der kuerzesten Kanten zurueckgeben
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    (cond
      ((equal? (length besucht) anzahl)  ; alle Knoten enthalten
       kuerzeste)
      ((null? priows)
       #f)
      (else
       'fehlt)
      ))                                 ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   )
  )
;;; Test geht nicht wegen der inneren Funktion
;;; ggf innere Funktion unabhaengig testen
;(prim-intern '(O F K P I) 5 '(Liste kuerzeste) '((B A 69)))
;(prim-intern '(O F K P I) 18 '(Liste kuerzeste) '((B A 69)))

; ---------------
; Schritt 4:
; Hier war man schon; es entstuende ein Zyklus (Hilfsfunktion)
;;; ===== zyklus? =====
; prüft, ob der freie Knoten der Kante in der Besuchtliste enthalten ist
( define (zyklus? kante besucht)
   (member (second kante) besucht))

; ----- Schritt 4 umgesetzt:
;weiter arbeiten ohne diese Kante in der Warteschlange
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    (cond
      ((equal? (length besucht) anzahl)  ; alle Knoten enthalten
       kuerzeste)
      ((null? priows)
       #f)
      ((zyklus? (first priows) besucht) ; freier Knoten in besucht?
       (prim-intern besucht anzahl kuerzeste (rest priows)))
      (else
       'fehlt)
      ))                                 ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   )
  )
;;; Test geht nicht wegen der inneren Funktion
;;; ggf innere Funktion unabhaengig testen
;(prim-intern '(O F K P I) 5 '(Liste kuerzeste) '((F O 14)))
;(prim-intern '(O F K P I) 18 '(Liste kuerzeste) '((B A 69)))

; ---------------
; Schritt 5:
; else: Der "Nomalfall", die Kante geht einzubauen,
; Liste besucht aktualisieren -> Hilfsfunktion
; aktuelle Kante in den Baum aufnehmen
; Nachfolger des freien Knoten in priows -> Hilfsfunktion

;;; ===== Hilfsfunktion aktualisiere-besuchtliste =====
(define
  (aktualisiere-besuchtliste knoten besucht)
  (cond
   ((member knoten besucht)
    besucht)
   (else
    (cons knoten besucht))
   ))

;; ===== Hilfsfunktion aktualisiere-priows =====
; Zur ersten Kante aus der Prioritaetswarteschlange werden alle weiter fuehrenden
; Kanten bestimmt und in den Rest der priows eingefuegt
(define
  (aktualisiere-priows priows)
  (fuege-alle-ein
   (nachfolge-kanten (second (first priows)) kanten)
   vor?
   (rest priows)))

; ----- Schritt 5 umgesetzt:
(define
  (prim start-knoten kanten)
  (define
    (prim-intern besucht anzahl kuerzeste priows)
    (cond
      ((equal? (length besucht) anzahl)  ; alle Knoten enthalten
       kuerzeste)
      ((null? priows)
       #f)
      ((zyklus? (first priows) besucht) ; freier Knoten in besucht?
       (prim-intern besucht anzahl kuerzeste (rest priows)))
      (else
       (prim-intern
        (aktualisiere-besuchtliste (second (first priows)) besucht)
        anzahl
        (cons (first priows) kuerzeste)
        (aktualisiere-priows priows)))
      ))                                 ; Ende interne Funktion
  (prim-intern                           ; Aufruf:der internen Funktion
   (list start-knoten)
   (anzahl-knoten kanten)
   '()
   (fuege-alle-ein ; Start-PrioWS
    (nachfolge-kanten start-knoten kanten)
    vor?
    '())
   )
  )

; Test:
(prim 'I kanten)